from .multi_area import MultiArea
from .multi_area_ei import MultiAreaEI
from .random_input import RandomInput

if __name__ == "__main__":
    print(MultiAreaEI)
    print(MultiArea)
    print(RandomInput)
